/*
 * logical - Logical (bitwise) operations on exact numbers
 *
 *  (c) 1999 Shiro Kawai  shiro@acm.org
 *  Terms and conditions to use, modify or distribute this code
 *  is the same as of STk.
 *
 *  $Id: logical.c,v 1.5 2000/02/17 11:05:42 shiro Exp $
 */

#include <stk.h>

/* A native implementation of logical operations for STk.
 * The interface is the same as SLIB's logical feature.
 * This implementation is much faster especially if the arguments fit
 * in the SMALLINT.
 */

/* 
 * It may be better if these functions are exported from number.c
 */
static SCM gimme_a_bignum()
{
    SCM z;
    MP_INT *m;
    STk_ignore_signals();
    NEWCELL(z, tc_bignum);
    m = (MP_INT*)must_malloc(sizeof(MP_INT));
    mpz_init(m);
    BIGNUM(z) = m;
    STk_allow_signals();
    return z;
}

static SCM integer2bignum(SCM i)
{
    SCM bignum;
    bignum = gimme_a_bignum();
    mpz_init_set_si(BIGNUM(bignum), INTEGER(i));
    return bignum;
}

/* If the result of evaluation falls in the range of INTEGER, convert it. */
static SCM normalize_integer(SCM num)
{
    long inum;
    if (BIGNUMP(num)) {
        inum = STk_integer_value(num);
        if (inum >= SMALLINT_MIN && inum <= SMALLINT_MAX) {
            return STk_makeinteger(inum);
        }
    }
    return num;
}

/* Architecture dependent parameters.
 * Currently I calculate them in the init routine.  Is there any reliable
 * way to obtain them from some header files?
 */
static int bits_in_smallint_max;
static int bits_in_ulong;

/*
 * PRIMITIVEs
 */

/* logand n1 n2 */
PRIMITIVE STk_logand(SCM n1, SCM n2)
{
    SCM result;
    static char *proc_name = "logand";
    if (!EXACTP(n1)) Serror("Exact number required", n1);
    if (!EXACTP(n2)) Serror("Exact number required", n2);
    if (INTEGERP(n1)) {
        if (INTEGERP(n2)) {
            return STk_makeunsigned(INTEGER(n1) & INTEGER(n2));
        }
        n1 = integer2bignum(n1);
    } else {
        if (INTEGERP(n2)) {
            n2 = integer2bignum(n2);
        }
    }
    result = gimme_a_bignum();
    mpz_and(BIGNUM(result), BIGNUM(n1), BIGNUM(n2));
    return normalize_integer(result);
}

/* logior n1 n2 */
PRIMITIVE STk_logior(SCM n1, SCM n2)
{
    SCM result;
    static char *proc_name = "logior";
    if (!EXACTP(n1)) Serror("Exact number required", n1);
    if (!EXACTP(n2)) Serror("Exact number required", n2);
    if (INTEGERP(n1)) {
        if (INTEGERP(n2)) {
            return STk_makeunsigned(INTEGER(n1) | INTEGER(n2));
        }
        n1 = integer2bignum(n1);
    } else {
        if (INTEGERP(n2)) {
            n2 = integer2bignum(n2);
        }
    }
    result = gimme_a_bignum();
    /* fgmp and gmp differ in this API */
#ifdef USE_FGMP
    mpz_or(BIGNUM(result), BIGNUM(n1), BIGNUM(n2));
#else
    mpz_ior(BIGNUM(result), BIGNUM(n1), BIGNUM(n2));
#endif
    return normalize_integer(result);
}

/* logxor n1 n2 */
PRIMITIVE STk_logxor(SCM n1, SCM n2)
{
    SCM result, tmp1, tmp2, tmp3;
    static char *proc_name = "logxor";
    if (!EXACTP(n1)) Serror("Exact number required", n1);
    if (!EXACTP(n2)) Serror("Exact number required", n2);
    if (INTEGERP(n1)) {
        if (INTEGERP(n2)) {
            return STk_makeunsigned(INTEGER(n1) ^ INTEGER(n2));
        }
        n1 = integer2bignum(n1);
    } else {
        if (INTEGERP(n2)) {
            n2 = integer2bignum(n2);
        }
    }
    result = gimme_a_bignum();
#ifdef USE_FGMP
    /* fgmp has xor function */
    mpz_xor(BIGNUM(result), BIGNUM(n1), BIGNUM(n2));
#else
    /* Somehow, gmp doesn't have xor function.
     * I use (!A & B) | (A & !B)
     */
    tmp1 = gimme_a_bignum();
    tmp2 = gimme_a_bignum();
    tmp3 = gimme_a_bignum();
    mpz_com(BIGNUM(tmp1), BIGNUM(n1));
    mpz_and(BIGNUM(tmp2), BIGNUM(tmp1), BIGNUM(n2));
    mpz_com(BIGNUM(tmp1), BIGNUM(n2));
    mpz_and(BIGNUM(tmp3), BIGNUM(tmp1), BIGNUM(n1));
    mpz_ior(BIGNUM(result), BIGNUM(tmp2), BIGNUM(tmp3));
#endif
    return normalize_integer(result);
}

/* lognot n1 */
PRIMITIVE STk_lognot(SCM n1)
{
    static char *proc_name = "lognot";
    if (INTEGERP(n1)) return STk_makeinteger(~INTEGER(n1));
    if (BIGNUMP(n1)) {
        SCM result = gimme_a_bignum();
        mpz_com(BIGNUM(result), BIGNUM(n1));
        return normalize_integer(result);
    }
    Serror("Exact number required", n1);
    return NIL;     /*NOTREACHED*/
}

/* logtest num1 num2 */
PRIMITIVE STk_logtest(SCM num1, SCM num2)
{
    static char *proc_name = "logtest";
    if (INTEGERP(num1) && INTEGERP(num2)) {
        return (INTEGER(num1) & INTEGER(num2)) ? Truth : Ntruth;
    }
    if (!EXACTP(num1)) Serror("exact number rqeuired", num1);
    if (!EXACTP(num2)) Serror("exact number rqeuired", num2);
    return (STk_zerop(STk_logand(num1, num2)) == Ntruth)? Truth : Ntruth;
}

/* logbit? index num */
PRIMITIVE STk_logbit(SCM index, SCM num)
{
    static char *proc_name = "logbit";
    long idx = STk_integer_value(index);
    if (idx == LONG_MIN || idx < 0) {
        Serror("index out of range", index);
    }
    if (INTEGERP(num)) {
        int inum = INTEGER(num);
        if (idx > bits_in_smallint_max) {
            return (inum < 0)? Truth : Ntruth;
        } else {
            return ((1L << idx) & inum)? Truth : Ntruth;
        }
    } else if (BIGNUMP(num)) {
        MP_INT r;
        int flag;
        int sign = mpz_cmp_si(BIGNUM(num), 0L);
        if (sign == 0) return STk_makeinteger(0); /* special case */
        STk_ignore_signals();
        mpz_init(&r);
        mpz_div_2exp(&r, BIGNUM(num), idx);
        flag = mpz_get_si(&r) & 0x01;
        mpz_clear(&r);
        STk_allow_signals();
        if (sign > 0) {
            return flag? Truth : Ntruth;
        } else {
            return flag? Ntruth : Truth;
        }
    }
    /*NOTREACHED*/
}

/* ash num count*/
PRIMITIVE STk_ash(SCM num, SCM cnt)
{
    static char *proc_name = "ash";
    long icnt;
    if (!EXACTP(cnt)) Serror("exact number required", cnt);
    icnt = STk_integer_value(cnt);
    if (icnt == LONG_MIN) {
        /* we don't support such large number of shift count */
        Serror("shift count overflow", cnt);
    }
    if (icnt == 0) return num;
    
    if (INTEGERP(num)) {
        long inum = INTEGER(num);
        if (icnt <= -bits_in_ulong) {
            /* we filter this condition out because native shift oprator
               overflows */
            inum = (inum < 0? -1 : 0);
            return STk_makeinteger(inum);
        } else if (icnt < 0) {
            if (inum < 0) {
                inum = ~((~inum) >> (-icnt));
            } else {
                inum >>= -icnt;
            }
            return STk_makeinteger(inum);
        } else if (icnt < bits_in_smallint_max) {
            /* need to check overflow */
            if (inum < 0) {
                if (-inum < (SMALLINT_MAX >> icnt)) {
                    inum <<= icnt;
                    return STk_makeinteger(inum);
                } 
            } else {
                if (inum < (SMALLINT_MAX >> icnt)) {
                    inum <<= icnt;
                    return STk_makeinteger(inum);
                } 
            }
        }
        /* Here, we know the result must be a bignum. */
        {
            SCM bnum = integer2bignum(num);
            SCM result = gimme_a_bignum();
            mpz_mul_2exp(BIGNUM(result), BIGNUM(bnum), icnt);
            return result;
        }
    } else if (BIGNUMP(num)) {
        SCM result = gimme_a_bignum();
        if (icnt < 0) {
            mpz_div_2exp(BIGNUM(result), BIGNUM(num), -icnt);
            if (mpz_cmp_si(BIGNUM(num), 0L) < 0
                && mpz_cmp_si(BIGNUM(result), 0L) == 0) {
                mpz_set_si(BIGNUM(result), -1L);
            }
        } else {
            mpz_mul_2exp(BIGNUM(result), BIGNUM(num), icnt);
        }
        return normalize_integer(result);
    }
    Serror("Exact number required", num);
    return Ntruth;/*NOTREACHED*/
}

/* copy-bit index num bool */
PRIMITIVE STk_copy_bit(SCM sindex, SCM snum, SCM sbool)
{
    static char *proc_name = "copy-bit";
    int index;
    if (!EXACTP(sindex)) Serror("exact number required", sindex);
    index = STk_integer_value(sindex);
    if (index < 0) Serror("index out of range", sindex);
    if (INTEGERP(snum) && index <= bits_in_smallint_max) {
        if (sbool == Ntruth) {
            return STk_makeinteger(INTEGER(snum)&(~(1L<<index)));
        } else {
            return STk_makeinteger(INTEGER(snum)|(1L<<index));
        }
    } else if (!EXACTP(snum)) {
        Serror("exact number requred", snum);
    }
    if (INTEGERP(snum)) snum = integer2bignum(snum);
    if (sbool == Ntruth) {
        SCM r = gimme_a_bignum();
        mpz_set(BIGNUM(r), BIGNUM(snum));
        mpz_clrbit(BIGNUM(r), (unsigned long)index);
        return r;
    } else {
        /* I don't know why gmp doesn't have mpz_setbit */
        MP_INT m, n;
        SCM r = gimme_a_bignum();
        STk_ignore_signals();
        mpz_init(&m);
        mpz_init(&n);
        mpz_set_ui(&m, 1);
        mpz_mul_2exp(&n, &m, index);
#ifdef USE_FGMP
        mpz_or(BIGNUM(r), BIGNUM(snum), &n);
#else
        mpz_ior(BIGNUM(r), BIGNUM(snum), &n);
#endif
        mpz_clear(&n);
        mpz_clear(&m);
        STk_allow_signals();
        return r;
    }
    /*NOTREACHED*/
}

/* integer-length num */
PRIMITIVE STk_integer_length(SCM snum)
{
    static char *proc_name = "integer-length";
    if (!EXACTP(snum)) Serror("exact number expected", snum);
    if (INTEGERP(snum)) {
        long num = INTEGER(snum);
        unsigned long mask = 1;
        unsigned long count = 0;
        if (num < 0) num = ~num;
        while (count < bits_in_smallint_max) {
            if ((num & (mask - 1)) == num) {
                return STk_makeinteger(count);
            }
            count++;
            mask<<=1;
        }
        return STk_makeinteger(count);
    } else {
        int sign = mpz_cmp_si(BIGNUM(snum), 0L);
        if (sign < 0) {
            SCM fix = gimme_a_bignum();
            mpz_add_ui(BIGNUM(fix), BIGNUM(snum), 1);
            return STk_makeinteger(mpz_sizeinbase(BIGNUM(fix), 2));
        } else {
            return STk_makeinteger(mpz_sizeinbase(BIGNUM(snum), 2));
        }
    }
}

/* bitmask n
   Returns (1 << n) - 1 */
PRIMITIVE STk_bitmask(SCM sindex)
{
    static char *proc_name = "bitmask";
    int index = STk_integer_value(sindex);
    if (index < 0) Serror("argument out of range", sindex);
    if (index <= bits_in_smallint_max) {
        unsigned long u = 1L << index;
        return STk_makeunsigned(u - 1);
    } else {
        SCM r = gimme_a_bignum();
        SCM tmp1 = gimme_a_bignum();
        mpz_set_ui(BIGNUM(r), 1);
        mpz_mul_2exp(BIGNUM(tmp1), BIGNUM(r), index);
        mpz_sub_ui(BIGNUM(r), BIGNUM(tmp1), 1);
        return r;
    }
    /*NOTREACHED*/
}


void STk_init_logical()
{
    int i;
    unsigned long l;
    
    /* calculate architecture dependent parameters */
    l = SMALLINT_MAX;
    for (i=0; l >= 1; i++)  l >>= 1;
    bits_in_smallint_max = i;
    l = ULONG_MAX;
    for (i=0; l >= 1; i++)  l >>= 1;
    bits_in_ulong = i;

    /* register primitives */
    STk_add_new_primitive("logand", tc_subr_2, (PRIMITIVE(*)()) STk_logand);
    STk_add_new_primitive("logior", tc_subr_2, (PRIMITIVE(*)()) STk_logior);
    STk_add_new_primitive("logxor", tc_subr_2, (PRIMITIVE(*)()) STk_logxor);
    STk_add_new_primitive("lognot", tc_subr_1, (PRIMITIVE(*)()) STk_lognot);
    STk_add_new_primitive("logtest", tc_subr_2,(PRIMITIVE(*)()) STk_logtest);
    STk_add_new_primitive("logbit?", tc_subr_2, (PRIMITIVE(*)()) STk_logbit);
    STk_add_new_primitive("ash", tc_subr_2, (PRIMITIVE(*)()) STk_ash);
    STk_add_new_primitive("integer-length", tc_subr_1, (PRIMITIVE(*)())STk_integer_length);
    STk_add_new_primitive("copy-bit", tc_subr_3, (PRIMITIVE(*)()) STk_copy_bit);
    STk_add_new_primitive("bitmask", tc_subr_1, (PRIMITIVE(*)()) STk_bitmask);
}

